#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>

#include "parser.h"


static struct link_url *links;
static struct link_url *links_tail;
static struct link_url *links_stack;

/*
 *	Report the top link
 */

struct link_url *link_top(void)
{
	return links_stack;
}

/*
 *	Add a box to a link area. Do merges to keep count down along lines
 */

void link_add_box(struct link_url *link, int x, int y, int w, int h)
{
	struct link_box *b;
	
	/*
	 *	Clip boxes partially off the top 
	 */
	 
	if(y<0)
		y = 0;
		
	/* Firstly try to merge chains of boxes on one line */
	if(link->boxes)
	{
		if(link->boxes->y == y && link->boxes->h == h)
		{
			if(x <= link->boxes->x + link->boxes->w &&
			   x >= link->boxes->x)
			{
				link->boxes->w = x - link->boxes->x + w;
				return;
			}
		}
	}
	/* Then add new boxes */
	b = (struct link_box *)malloc(sizeof(struct link_box));
	if(b==NULL)
		parser_fatal("Out of memory");
	b->next = link->boxes;
	b->x = x;
	b->y = y;
	b->w = w;
	b->h = h;
	link->boxes = b;
}

/*
 *	Flush the link areas
 */	
 
void link_free(void)
{
	struct link_url *l;
	struct link_box *b;

	while(links!=NULL)
	{
		l = links;
		links = links->link;
		while(l->boxes)
		{
			b=l->boxes;
			l->boxes = b->next;
			free(b);
		}
		free(l);
	}
	links = links_tail = links_stack = NULL;
}

/*
 *	Find matching link
 */			
 
struct link_url *link_hit(int x, int y)
{
	struct link_url *l = links;
	struct link_box *b;

	while(l!=NULL)
	{
		b = l->boxes;
		while(b!=NULL)
		{
			if(x >= b->x && x < b->x+b->w &&
			   y >= b->y && y < b->y+b->h)
			   	return l;
			b=b->next;
		}
		l=l->link;
	}
	return NULL;
}

void link_area_begin(struct token_stack *t, int type, char *uri)
{
	struct link_url *l = (struct link_url *)malloc(sizeof(struct link_url));
	if(l==NULL)
		parser_fatal("Out of memory");
	l->token = t;
	l->type = type;
	l->data = uri;
	l->boxes = NULL;

	l->link = NULL;
	if(links==NULL)
		links = l;
	else
		links_tail->link = l;	
	links_tail = l;

	l->next = links_stack;
	links_stack = l;
}

void link_area_end(struct token_stack *t, int type)
{
	while(links_stack)
	{
		if(links_stack->type == type)
		{
			links_stack=links_stack->next;
			return;
		}
		links_stack = links_stack->next;
	}
}

